log4js = require 'log4js'
logger = log4js.getLogger 'TaskMaster'

Task = require './Task.coffee'

createTaskModel = require '../models/Task.coffee'

EventEmitter = require 'events'

class TaskMasterEmitter extends EventEmitter

# options.newTaskEventName == 'bot.123123123.newtask'
# options.tableName == 'bot_123123123_tasks'
# options.maxTasksCount == 1
class TaskMaster
  constructor: (@interface, @mongoose = null, @options = {}) ->
    @emitter = new TaskMasterEmitter

    unless @options
      logger.warn 'options required'
      return

    @activeTasks = []
    @updatingTasks = false

    @taskSubscribtions = {}

    @interface.subscribe @options.newTaskEventName, =>
      unless @inprocess() then @checkTasks()

TaskMaster::initListeners = ->
  @emitter.on 'remoteTaskResult', (task) =>
    @emitter.emit "task.#{task.id}.result", task
    if task.state == 'completed' then @emitter.emit "task.#{task.id}.completed", task
    if task.state == 'failed' then @emitter.emit "task.#{task.id}.failed", task

TaskMaster::inprocess = ->
  @activeTasks.length >= @options.maxTasksCount or @checkingTasks

TaskMaster::checkTasks = ->
  if @checkingTasks then return Promise.resolve()
  @checkingTasks = true
  if @options.maxTasksCount - @activeTasks.length == 0
    @checkingTasks = false
    return Promise.resolve()
  Model = createTaskModel @options.tableName, @mongoose
  req = Model.find state: 'created'
  if isFinite @options.maxTasksCount
    req = req.limit @options.maxTasksCount - @activeTasks.length
  req.exec()
  .then (docs) =>
    docs.forEach (doc) =>
      doc = doc.toObject()
      _task = Task.parse doc
      if _task
        @activeTasks.push _task
        @handleTask _task
    @checkingTasks = false
    Promise.resolve @activeTasks
  , (err) =>
    @checkingTasks = false
    Promise.reject err

TaskMaster::handleTask = (task) ->
  @emitter.emit "task.#{task.action}", task
  @emitter.emit "newTask", task

  completeHandler = (data) =>
    @updateTask task
    @removeActiveTask task
    @interface.publish task.taskResultChannelName, task.info()
    unless @inprocess() then @checkTasks()

  task.emitter.on 'complete', completeHandler
  task.emitter.on 'error', completeHandler

  task.emitter.on 'complete', =>
    @emitter.emit 'taskCompleted', task
  task.emitter.on 'error', =>
    @emitter.emit 'taskFailed', task

  task.emitter.on 'step', (data) =>
    @emitter.emit 'taskStep',
      task: task
      data: data

TaskMaster::updateTask = (task) ->
  Model = createTaskModel @options.tableName, @mongoose
  req = Model.update
    taskId: task.id
  , task.info()

  req.exec()
  .then (doc) =>
    Promise.resolve doc
  , (err) =>
    Promise.reject err

TaskMaster::removeActiveTask = (task) ->
  @activeTasks.splice @activeTasks.indexOf(task), 1

TaskMaster::publishTask = (tableName, eventChannel, task) ->
  _Model = createTaskModel tableName, @mongoose
  _task = new _Model task.info(), false
  _task.save()
  .then (doc) =>
    @interface.publish eventChannel, task.id
    @onceSubscribeTask task
    Promise.resolve doc

TaskMaster::onceSubscribeTask = (task) ->
  if @taskSubscribtions[task.taskResultChannelName] then return
  @interface.subscribe task.taskResultChannelName, (data) => @taskResultHandler(data)

TaskMaster::taskResultHandler = (data) ->
  task = Task.parse data
  if task
    @emitter.emit 'remoteTaskResult', task
    if task.result
      @emitter.emit 'remoteTaskCompleted', task
    else
      @emitter.emit 'remoteTaskFailed', task


module.exports = TaskMaster